#!/usr/bin/python

#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: argparser.py,v 1.7 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This provides the ArgumentParser class which parses X{command argument}s
automatically into a dictionary as well as a series of Parser classes
which handle single arguments and TypeCheckers which handle validating
user input values and converting them into other types (string -> int...).

It's very complex, however, it allows us to nicely centralize a lot
of the duties involved in parsing user input into one place rather than
doing it in every single command over and over again.  This way commands
can specify their syntax line and the ArgParser handles the rest returning
to the commands a dict with the name/value pairs including argument 
defaults.

When building an arg spec, you can pass into the ArgumentParser several
options which change the way it parses the user input.

Supported options::

  stripBraces (default=on): whether all arguments should have 
            braces stripped before being parsed

  noparsing (default=off): doesn't insure that all arguments are
            parsed.  works well when matched with limitparsing=0 to 
            provide a syntax line for commands that parse their own 
            input

  limitparsing:int (default=-1): only parse this number of tokens 
            into dict, the rest of the input line goes into 
            argdict["input"]

  nodefaults:bool (default=off): turn off default lookups through variables.

Refer to the modules.lyntincmds and modules.tintincmds for examples
on arg specs and commands and how it all intertwines.
"""
import re, time
from lyntin import utils

defaultOptions={ "stripBraces": 1, "noparsing": 0, "limitparsing": -1, "nodefaults":0 }
optionParser = None

class ParserException(Exception):
  pass

class ArgumentParser:
  """
  This is the actual ArgumentParser class.  Each command gets its
  own ArgumentParser.  It handles taking in user input and parsing
  it into a series of arguments which are then placed in a dict and
  passed to the command function.  This centralizes parsing user
  input into one place and also enables us to centralize type checking,
  argument checking, and conversion in one place as well.
  """
  def __init__(self, argspec, argoptions=None):
    """
    Initializes the ArgParser instance so it can then parse user
    input for arguments.  It also generates the syntaxline which
    is used by the help system to tell the user what the command
    syntax is.

    @param argspec: the argument specification for the command
    @type  argspec: string

    @param argoptions: the options to use to parse the argspec
    @type  argoptions: string
    """
    # the syntax line is automatically generated from the argspec.
    # we print it out whenever we have a ParserException in the user input.
    self.syntaxline = ""

    if argoptions:
      self.buildOptions(argoptions)
    else:
      self.options = defaultOptions.copy()
    self.buildParsers(argspec)
    return

  def getOption(self, optionname):
    """
    Gets a specific option from the arg options hash as passed into the
    __init__.

    @param optionname: the name of the option
    @type  optionname: string

    @return: the value of the option or None
    @rtype: varies (string or int)
    """
    if optionname in self.options:
      return self.options[optionname]
    else:
      return None

  def buildOptions(self, argoptions):
    """
    Set the options for this ArgumentParser
    key=value values in argoptions get put directly into self.options
    other values in argoptions get set to 1 in self.options.

    example 1::

      argoptions="ignorall loud:boolean=true"
      self.options={"ignorall":1,"loud":1}

    example 2::

      argoptions="lalala wewewewe=hahaha"
      self.options={"lalala":1,"wewewewe":"hahaha"}

    @param argoptions: the argoptions as passed into __init__.
    @type  argoptions: string
    """
    # FIXME - this should operate on a locally created dict and pass
    # that dict back to __init__ which should set self.options and
    # self.argoptions
    global optionParser
    self.argoptions=argoptions
    self.options=defaultOptions.copy()

    if optionParser == None:
      optionParser = ArgumentParser("otherOptions* otherValuedOptions**")
    argdict = optionParser.parse(argoptions)

    for key in argdict:
      if key == "otherOptions":
        for otherOption in argdict[key]:
          self.options[otherOption] = 1
          if len(otherOption)>3 and otherOption[0:2] == "no":
            self.options[otherOption[2:]] = 0
      elif key == "otherValuedOptions":
        self.options.update(argdict[key])
      else:
        self.options[key] = argdict[key]

    # set types for certain options
    self.options["limitparsing"] = int(self.options["limitparsing"])

  def buildParsers(self, argspec):
    """
    Build up the set of parsers to be used for argument parsing.

    The argspec follows the following format:

      1. [argname[:argtype]]+ 
      2. [argname[:argtype]=defaultval]+ 
      3. [argname:argtype*] 
      4. [argname[:argtype]]+ 
      5. [argname[:argtype]=defaultval]+ 
      6. [argname:argtype**]

    Any of the arguments can be specified either by name or populated
    by position, except for arguments after the index collector
    argument.  Those must be specified by name only.

    Once one default value is given all further arguments must have
    default values (except collector arguments, which have implicit
    default arguments of the empty list and the empty map).

    For examples see the test code at the end of argparser.py.

    @param argspec: the argument specification to use to parse future
        user input
    @type  argspec: string
    """
    self.parsers = {}
    self.indexparsers = []
    self.extraindexparser = None
    self.extranamedparser = None

    self.argspec = self.split(argspec, buildsyntaxline=1)

    parsedspec = self.argspec

    doneWithIndices = 0
    defaultSeen = 0
    for i in range(0,len(parsedspec)):
      namedCollector = 0
      indexCollector = 0
      argname, argdef = parsedspec[i]
      if argname.find(":") > -1:
        argname,typespec = argname.split(":",1)
      else:
        # extra argname assignment here is just for consistency
        argname, typespec = argname, "string"

      if len(argname) >= 1 and argname.endswith("*"):
        if argdef != None:
          raise ParserException("cannot specify a default value for a collection argument (%s=%s)" % (argname, argdef))

        if len(argname) >= 2 and argname.endswith("**"):
          argname = argname[:-2]
          if i < len(parsedspec) -1:
            raise ParserException("named collection argument must be the last argument (%s)" % (argname))
          parser = ExtraNamedParser(self, argname)
          namedCollector = 1

        else: #this is an index collection argument
          argname = argname[:-1]
          parser = ExtraIndexParser(self, argname)
          indexCollector = 1
          doneWithIndices = 1
          
      else:
        parser = Parser(self,argname)

      typechecker = _createTypeChecker(typespec)
      if not typechecker:
        raise ParserException("Unknown type specifier: %s" % (typespec))

      parser.setTypeChecker(typechecker)

      if argdef != None:
        parser.setDefault(parser.parse(argdef))
        defaultSeen = 1

      if defaultSeen and not parser.defaultset:
        raise ParserException("Argument without default value (%s) seen after default values already specified" % (argname))
      
      if not namedCollector and not indexCollector:
        if not doneWithIndices:
          self.indexparsers.append(parser)
        if argname in self.parsers:
          raise ParserException("Multiple argument named %s specified." % (argname))
        self.parsers[argname] = parser
      elif namedCollector:
        self.extranamedparser = parser
        self.parsers[argname] = parser
      elif indexCollector:
        self.extraindexparser = parser
        self.parsers[argname] = parser
    
  def parse(self, input, defaultresolver=None):
    """
    Takes an input string and produces the populated dictionary
    matching self's argspec.  

    @param input: the user input string
    @type  input: string

    @param defaultresolver: A function that will take an argument 
        name and return a default value (None if there should be none) 
        to override builtin defaults.
    @type  defaultresolver: function

    @return: the populated dictionary of all the args and values
    @rtype: dict

    @raise ParserException: if extra arguments are encountered without
        appropriate collection arguments specified, or if required arguments
        are missing, or if arguments passed in aren't valid
    """    
    argdict = {}

    arguments = self.split(input, self.getOption("limitparsing"))

    foundNamedArg = 0
    for i in range(0,len(arguments)):
        key,val = arguments[i]

        if val == None:
          if foundNamedArg and not self.extraindexparser:
            raise ParserException("Non-named argument (%s) found after Named argument" % (key))
          if i < len(self.indexparsers):
            parser = self.indexparsers[i]
          elif self.extraindexparser:
            parser = self.extraindexparser
          else:
            raise ParserException(f"Unexpected argument received {key} by the argument parser.")
          parser.parseInto(i, key, argdict)
        else:
          foundNamedArg = 1
          if key in self.parsers:
            parser = self.parsers[key]
          else:
            matchedkeys = [arg for arg in self.parsers if arg.startswith(key)]

            if len(matchedkeys) == 1:
              parser = self.parsers[matchedkeys[0]]
            elif self.extranamedparser:
              parser = self.extranamedparser
            else:
              if len(matchedkeys) == 0:
                raise ParserException("Invalid named argument: %s=%s" % (key,val))
              else:
                raise ParserException("Ambiguous named argument: %s=%s %s" % (key, val, matchedkeys))
            
          parser.parseInto(key, val, argdict)


    # now check that everything has been specified, putting in defaults 
    # where available
    for key in self.parsers:
      if key not in argdict:
        # gotta be careful here with the extra defaultset value since
        # the parser may parse a string into None, or anything really
        default = None
        defaultset = 0
        parser = self.parsers[key]

        if not self.getOption("nodefaults") and defaultresolver:
          default = defaultresolver(key)
          if default != None:
            default = parser.parse(default)
            defaultset = 1
            
        if not defaultset and parser.defaultset:
          default = parser.default
          defaultset = 1
        
        if not defaultset and not self.getOption("noparsing"):
          raise ParserException("Must specify a value for argument %s" % (key))
        else:
          argdict[key] = default
          
    return argdict

  def split(self, input, maxsplit=-1, buildsyntaxline=0):
    """
    Take an input string and tokenizes it into a list of pairs.
    Tokens with equal signs come back as (key,value) pairs, those
    without come back as (argument,None)

    {}s are treated like quotes, and everything between the {}s is
    ignored.

    Any amount of white space between arguments is ignored.  (No empty
    arguments are returned.)

    "\\" (single backslash) escapes anything, including =, {, }, 
    \\ (single backslash), and any character, so "\\a" (single 
    backslash then a) becomes "a" in the argument, and
    "\\n\\o\\t\\a\\d\\r\\a\\g\\o\\n" (a series of single backslashes 
    followed by a letter) is the same as "notadragon".

    After maxplit arguments are parsed (all of the arguments if maxsplit 
    < 0), we'll stop and return the rest of input as the final item.

    @param input: the user input to tokenize
    @type  input: string

    @param maxsplit: the maximum number of pairs to split
    @type  maxsplit: int (-1 if no maxplit)

    @param buildsyntaxline: whether or not to build the syntax
        line as we're going along (this method is used both when
        the ArgParser is initialized as well as for parsing
        user input)
    @type  buildsyntaxline: int (0 or 1)

    @return: the split input
    @rtype: list of 2-length-tuples

    @raises ParserException: if \\ (single backslash) is found at the end 
        of the line or if mismatched { or } are found
    """
    bracketdepth = 0
    arg = ""
    val = None
    equalsign = 0
    arguments = []
    while input and (maxsplit < 0 or len(arguments) < maxsplit):
      nextchar = input[0:1]
      input = input[1:]

      if nextchar == " " or nextchar == "\t":
        if not bracketdepth:
          # We've completed a full argument
          if arg!="":
            arguments.append( (arg,val) )
            if buildsyntaxline:
              synarg = arg.upper()
              if synarg.endswith("*"):
                synarg = synarg[:-1] + "..."

              if val and len(val) > 0:
                synarg = synarg + "=" + val

              if equalsign == 1:
                self.syntaxline += "[<%s>] " % synarg
              else:
                self.syntaxline += "[%s] " % synarg

          arg = ""
          val = None
          equalsign = 0
        else:
          if val != None:
            val = val + nextchar
          else:
            arg = arg + nextchar
      elif nextchar == "\\":
        if input == "":
          raise ParserException("\\ at end of line.")
        else:
          nextchar = input[0:1]
          input = input[1:]
          if val != None:
            val = val + nextchar
          else:
            arg = arg + nextchar
      elif nextchar == "}":
        bracketdepth = bracketdepth - 1
        if bracketdepth < 0:
          raise ParserException("mismatched }")
        if val != None:
          val = val + nextchar
        else:
          arg = arg + nextchar
      elif nextchar == "{":
        bracketdepth = bracketdepth + 1
        if val != None:
          val = val + nextchar
        else:
          arg = arg + nextchar
      elif val == None and bracketdepth == 0 and nextchar == "=":
        val = ""
        equalsign = 1
      else:
        if val != None:
          val = val + nextchar
        else:
          arg = arg + nextchar

    if bracketdepth:
      raise ParserException("Mismatched {")

    if arg != "":
      arguments.append( (arg, val) )
      if buildsyntaxline:
        synarg = arg.upper()
        if synarg.endswith("*"):
          synarg = synarg[:-1] + "..."

        if val and len(val) > 0:
          synarg = synarg + "=" + val

        if equalsign == 1:
          self.syntaxline += "[<%s>] " % synarg
        else:
          self.syntaxline += "[%s] " % synarg

      arg = ""
      val = ""

    if input:
      arguments.append( ("input", input) )
    
    return arguments

class Parser:
  """
  This is the base class for the parsers that argumentparser uses to
  actually populate the dictionary with each argument.
  """
  def __init__(self, argparser, argname):    
    """
    Initializes the Parser.

    @param argparser: the argparser instance this belongs to
    @type  argparser: argparser.ArgParser

    @param argname: the name of the argument this parser handles type
        and value checking for.
    @type  argname: string
    """
    self.argname = argname
    self.default = None
    self.defaultset = 0
    self.typechecker = None
    self.argparser = argparser

  def setTypeChecker(self, typechecker):
    """
    Sets the typechecker.  The Parser then uses this to check the
    values coming in before setting them on the name/value dict.

    @param typechecker: the typechecker to set
    @type  typechecker: argparser.TypeChecker
    """
    self.typechecker = typechecker

  def parseInto(self, key, val, argdict):
    """
    Populates the argument dictionary with the appropriate value.

    @param key: the argument name
    @type  key: string

    @param val: the argument value
    @type  val: string

    @param argdict: the argument dictionary to populate
    @type  argdict: dict

    @raises ParserException: if multiple values were given for the argument
    """
    if self.argname in argdict:
      raise ParserException("Multiple values for argument %s given" % (self.argname))
    else:
      argdict[self.argname] = self.parse(val)

  def parse(self, val):
    """
    Parses the value according to this parser and its associated 
    Typechecker.  We both typecheck it here (making sure it's valid)
    as well as convert it (i.e. strings -> ints, strings -> regexps...).

    @param val: the value to parse
    @type  val: string

    @return: the newly adjusted value
    @rtype: varies
    """
    if self.argparser.getOption("stripBraces"):
      val = utils.strip_braces(val)
    if self.typechecker:
      return self.typechecker.check(val)
    else:
      return val

  def setDefault(self, val):
    """
    Sets the default value for this Parser which is used when no
    other value is given.  Also sets the "defaultset" member which
    tells the Parser that it has a default value.

    @param val: the default value
    @type  val: string or int
    """
    self.default = val
    self.defaultset = 1
      
class ExtraIndexParser(Parser):
  """
  This class captures the parsing behaviour for an index collector.
  for each call to parseInto an entry is put into the list value in
  the argument dictionary.
  """
  def __init__(self, argparser, argname):
    """
    Initializes the Parser.

    @param argparser: the argparser instance this belongs to
    @type  argparser: argparser.ArgParser

    @param argname: the name of the argument this parser handles type
        and value checking for.
    @type  argname: string
    """
    Parser.__init__(self,argparser,argname)
    self.default = []
    self.defaultset = 1
    
  def parseInto(self, key, val, argdict):
    """
    Populates the argument dictionary with the appropriate value.

    @param key: the argument name
    @type  key: string

    @param val: the argument value
    @type  val: string

    @param argdict: the argument dictionary to populate
    @type  argdict: dict

    @raises ParserException: if multiple values were given for the argument
    """
    val = self.parse(val)
    if self.argname in argdict:
      argdict[self.argname].append(val)
    else:
      argdict[self.argname] = [val]

class ExtraNamedParser(Parser):
  """
  This class captures the parsing behaviour for a named value collector.
  for each call to parseInto a new key=value pair is put into a map
  in the argument dictionary.
  """
  def __init__(self,argparser,argname):
    """
    Initializes the Parser.

    @param argparser: the argparser instance this belongs to
    @type  argparser: argparser.ArgParser

    @param argname: the name of the argument this parser handles type
        and value checking for.
    @type  argname: string
    """
    Parser.__init__(self,argparser,argname)
    self.default = {}
    self.defaultset = 1
    
  def parseInto(self, key, val, argdict):
    """
    Populates the argument dictionary with the appropriate value.

    @param key: the argument name
    @type  key: string

    @param val: the argument value
    @type  val: string

    @param argdict: the argument dictionary to populate
    @type  argdict: dict

    @raises ParserException: if multiple values were given for the argument
    """
    val=self.parse(val)
    if self.argname in argdict:
      if key in argdict or key in argdict[self.argname]:
        raise ParserException("Multiple values given for argument %s" % (key))
      argdict[self.argname][key] = (val)
    else:
      argdict[self.argname] = {key:val}


typecheckers = {}

def _createTypeChecker(typespec):
  """
  Factory method for instantiating TypeChecker subclasses based on the
  type being passed in.

  First the typespec is split at its first colon.

  The first element (the typename) of the typespec is used as a key
  into typecheckers to find the function/class object to call to
  create the typechecker. 

  The rest of the typespec (the typeargs) is used with the function to
  construct the typechecker desired.

  @param typespec: the type that we're checking.  i.e. "string", "int", ...
      This translates directly into a specific TypeChecker.
  @type  typespec: string

  @return: the TypeChecker subclass created
  @rtype: argparser.TypeChecker
  """
  typespec = typespec.split(":",1)
  if len(typespec) == 1:
    typename, typeargs = typespec[0],None
  else:
    typename, typeargs = typespec

  if typename not in typecheckers:
    return None
  typechecker = typecheckers[typename](typename,typeargs)

  return typechecker

class TypeChecker:
  """
  Trivial base class for argument checkers
  """
  def __init__(self, typename, typeargs):
    """
    Initializes the TypeChecker.  Over-ridden by all the TypeChecker 
    subclasses.

    @param typename: the name of the type
    @type  typename: string

    @param typeargs: the arguments passed to the TypeChecker to initialize
        it
    @type  typeargs: tuple
    """
    return

  def check(self, arg):
    """
    Over-ridden by all the TypeChecker subclasses.

    @param arg: the argument to check the type-hood and convert
    @type  arg: string

    @returns: the converted argument
    @rtype: varies
    """
    return arg

class StringChecker(TypeChecker):
  """
  Essentiallly the same as the trivial base class, but it's explicit
  that we just return the string we take in without any transformation.
  """
  def __init__(self, typename, typeargs):
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))
    return

  def check(self, arg):
    """
    Returns the string we're looking at without any transformation.

    @param arg: the argument to pass back
    @type  arg: string

    @return: the argument
    @rtype:  string
    """
    return arg

typecheckers["string"] = StringChecker

class IntChecker(TypeChecker):
  """
  Accept only integer values and return integer objects.
  """
  def __init__(self, typename, typeargs):
    """
    Initializes the TypeChecker.  Over-ridden by all the TypeChecker 
    subclasses.

    @param typename: the name of the type
    @type  typename: string

    @param typeargs: the arguments passed to the TypeChecker to initialize
        it
    @type  typeargs: tuple

    @raises ParserException: If typeargs are passed in--this is 
        non-configurable.
    """
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))

  def check(self,arg):
    """
    Verifies the arg is of type int and converts it to an int.

    @param arg: the argument to check the type-hood and convert
    @type  arg: string

    @returns: the converted argument from string to int
    @rtype: int
    """
    return int(arg)

typecheckers["int"] = IntChecker

class BooleanChecker(TypeChecker):
  """
  Accept only boolean values using the utils.convert_boolean function.
  """
  def __init__(self, typename, typeargs):
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))

  def check(self,arg):
    """
    Verfies the argument is a boolean (using utils.convert_boolean function)
    and converts it to a boolean value (0 or 1).

    @param arg: the argument to check the type-hood and convert
    @type  arg: string

    @returns: the converted argument
    @rtype: int (0 or 1)

    @raises ParserException: if the arg is not a valid boolean
    """
    ret = utils.convert_boolean(arg)
    if ret == 1 or ret == 0:
      return ret

    raise ParserException("Invalid boolean value specified: %s" % (arg))

typecheckers["boolean"] = BooleanChecker

class BooleanOrNoneChecker(TypeChecker):
  """
  Accept only boolean values or special "not specified" values.  Booleans
  are handled by utils.convert_boolean.  Values not handled by that
  function will throw an exception.
  """
  def __init__(self, typename, typeargs):
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))

  def check(self,arg):
    """
    Verfies the argument is a boolean (using utils.convert_boolean function)
    and converts it to a boolean value (0 or 1).  Also handles None, "-"
    and the empty string as None.

    @param arg: the argument to check the type-hood and convert
    @type  arg: string

    @returns: the converted argument
    @rtype: int (0 or 1) or None

    @raises ParserException: if the arg is not a valid boolean
    """
    ret = utils.convert_boolean(arg)
    if ret == 1 or ret == 0:
      return ret
    elif arg == "None" or arg == "-" or arg == "":
      return None
    else:
      raise ParserException("Invalid boolean value specified: %s" % (arg))

typecheckers["booleanornone"] = BooleanOrNoneChecker

class EvalChecker(TypeChecker):
  """
  Evaluate its input argument as python code and return the resulting object.
  """
  def __init__(self, typename, typeargs):
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))

  def check(self,arg):
    """
    Assumes the argument is valid python code.  It evaluates the code
    and returns the result.

    @param arg: the argument to evaluate
    @type  arg: string

    @returns: the result of eval(arg)
    @rtype: varies

    @raises ParserException: if we can't eval the argument
    """
    try:
      return eval(arg)
    except Exception as e:
      raise ParserException("Error eval-ing argument (%s): %s" % (arg, e))

typecheckers["eval"] = EvalChecker 

class TimeSpanChecker(TypeChecker):
  """
  Accepts an amount of time and converts it to a number of seconds.
  """
  def __init__(self, typename, typeargs):
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))

  def check(self, arg):
    """
    Assumes the argument is a time span.

    @param arg: the timespan
    @type  arg: string

    @returns: the number of seconds in the timespan
    @rtype: int

    @raise ParserException: if the timespan is invalid
    """
    try:
      time = utils.parse_timespan(arg)
      return time
    except:
      raise ParserException("Invalid timespan specified %s" % (arg,))

typecheckers["timespan"] = TimeSpanChecker
  
class TimeChecker(TypeChecker):
  """
  Accepts a date specification.

  Will also accept a time specification and apply it as a delta from
  _now_.  converts to the standard seconds-from_epoch. 
  """
  def __init__(self, typename, typeargs):
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))

  def check(self, arg):
    """
    Assumes the argument is valid python code.  It evaluates the code
    and returns the result.

    @param arg: the time argument
    @type  arg: string

    @returns: the number of seconds since the epoch
    @rtype: int

    @raise ParserException: if the arg is not a valid time
    """
    time = utils.parse_time(arg)
    if time != None:
      return time
    else:
      raise ParserException("Invalid time specified %s" % (arg,))

typecheckers["time"] = TimeChecker
  
class ChoiceChecker(TypeChecker):
  """
  Allows for a value to come from a selection of different strings.
  Automatically expands to one of them if it is uniquely specified.

  Typeargs should be a |-delimitted list of possibly values
  """
  def __init__(self, typename, typeargs):
    """
    Uses the typeargs to set the choices availble.
    """
    if not typeargs:
      raise ParserException("TypeArgs (%s) not specified for %s type - must allow at least one choice." % (typeargs, typename) )
    self._choices = typeargs.split("|")

  def check(self, arg):
    """
    Checks the arg against a series of choices that were presented
    when this checker was instantiated.  So you could say that valid
    values of this argument are "high" and "low" and if the argument
    value the user passed in was "blue", then we would raise a 
    ParserException since that's not a valid value.

    @param arg: the argument
    @type  arg: string

    @raise ParserException: if the arg is not a valid choice
    """
    possibilities = []
    for item in self._choices:
      if item.find(arg) == 0:
        possibilities.append(item)
        if len(item) == len(arg):
          return item
    if len(possibilities) == 0 or len(possibilities) > 1:
      raise ParserException("Invalid argument, must be one of %s." % (self._choices,))
    else:
      return possibilities[0]

typecheckers["choice"] = ChoiceChecker

class ReChecker(TypeChecker):
  """
  Compiles the incoming argument as a regular expression.
  """
  def __init__(self, typename, typeargs):
    if typeargs:
      raise ParserException("TypeArgs (%s) specified for non-configurable type (%s)" % (typeargs, typename))

  def check(self, arg):
    """
    Converts the string into a regular expression.  Raises whatever
    exceptions re.compile raises.

    @param arg: the argument
    @type  arg: string

    @return: the compiled regular expression
    @rtype: Re
    """
    return re.compile(arg)

typecheckers["re"] = ReChecker

if __name__ == '__main__':
  testargs = {
    ("arg1 arg2 arg3* arg4**",None):["test1 test3 test5 test7 help=wahoo woo=weewee"],
    ("mapname*",None):["3k mapper by notadragon","lalala"],
    ("mapname*","noparsing"):["3k mapper by notadragon"],
    ("option* quiet:boolean=true",None):["a b c quiet=false d","a b c quiet=true","x b c"]} 

  for argspec,argoptions in testargs.items():
    argparser = ArgumentParser(argspec,argoptions)
    print(f"Argspec: {argspec}")
    if argoptions: print(f"Argopts: {argoptions}")
    for args in testargs[(argspec,argoptions)]:
      print(f"Args   : {args}")
      print(f"Dict   : {argparser.parse(args)}")

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: manual.py,v 1.5 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module holds the README manual text.
"""
from lyntin import exported

bugs = """
Lyntin was originally written by Lyn Headley.  Lyntin is currently 
being maintained by Will Guaraldi (willhelm@users.sourceforge.net) 
as of 1.3.2.

We appreciate ALL types of feedback.

Inevitably you will either run across a bug in Lyntin or the need 
for a feature to be implemented.  When this happens, we ask you 
to provide as much information as you can:

  - operating system, version of Python and version of Lyntin
    (from #diagnostics)
  - stacktrace (if it's a bug and kicked up a stacktrace)
  - explanation of what happened vs. what should be happening
  - any other pertinent information

Enter this in the bugs forum or send it to the mailing list.  
Details for both are on the Lyntin web-site:

   http://lyntin.sourceforge.net/

category: readme
"""

command = """
Lyntin uses Lyntin commands to allow you to manipulate the Lyntin 
client and setup your session with aliases, variables, actions, 
and such.  Commands start with the command character--by default 
this is "#".  It can be changed with the "#config" command.  The 
command character can also do some other special things:

1. You can execute commands in another session by typing the 
   command character and then the sesion name then the command.
   example:

      #3k say hello       - will say hello in session 3k
      #a #info            - will run the #info command in session a

2. You can switch to another session by typing the command 
   character and then the session name.  examples:

      #a                  - will switch to session a (if it exists)
      #3k                 - will switch to session 3k (if it exists)

3. You can execute a command in all sessions by typing the 
   command character then all.  examples:

      #all say hello      - will run "say hello" in all sessions

4. You can execute a command a number of times by typing the 
   command character then a number, then the command.  examples:

      #5 say hello        - will run "say hello" 5 times
      

Commands are separated by the semicolon.  Semicolons can be 
escaped with the \ character.  examples:

   say hello;wave         - will run "say hello" then "wave"
   say hi!  \;)           - will run "say hi!  ;)"

Command arguments can be enclosed with { }.  This enables you to 
specify arguments that have multiple words in them.  exaples:

   #alias a b             - executes #alias with args "a" and "b"
   #alias {a} {b}         - executes #alias with args "a" and "b"
   #alias {a} {say hi}    - executes #alias with args "a" and "say hi"
   #alias a say hi        - executes #alias with args "a", "say", 
                            and "hi" which will kick up an error
                            (since the #alias command doesn't accept
                            a third string argument)


{, } and \ can all be escaped with the \ character: \{, \}, and \\.

category: readme
"""

contribute = """
Development and maintenance is entirely managed by the maintainer 
right now.  If you're interested in sending in bug fixes, please 
feel welcome.  I'm extremely accomodating to most methods of 
sending in patches, however a diff -u against the code in cvs is
great for this sort of thing.

All patches and such things should be sent to the mailing list at:

   lyntin-devl@lists.sourceforge.net

NOTE: Patches will not be applied unless the author either yields 
the copyright to the FSF or to me (and then I'll yield it to the 
FSF).

Please read through the web-site areas on coding conventions and 
such before sending in code.

category: readme
"""

errata = """
The latest release of Lyntin is always available from:

   http://lyntin.sourceforge.net/

in the "download" section.  We also have snapshots of what's in CVS 
in the "development" section.

There are very few developers working on Lyntin, but we turn
around emails in very short periods of time.  Issues/bugs are
tracked by me by hand since the Sourceforge trackers _suck_.

When communicating bugs or feature requests INCLUDE AS MUCH 
INFORMATION AS POSSIBLE.  If you don't, I'll just ask you for
it anyhow.

In-game help is accessed by typing ``#help``.  When you start 
Lyntin for the first time, type ``#help general`` and that'll 
get you started.  Read through the various help files at your 
leisure.

category: readme
"""

evaluation = """
Variables get evaluated according to a new methodology we developed
after disliking the Tintin way of doing things.  This new methodology
provides maximum flexibility and provides for some things that
the Tintin variable evaluation did not provide for.

  * Variables and placement variables with a single $ or % are
    evaluated/expanded.  One layer of $s or %s is then stripped 
    off when this evaluation occurs.

  * Variables are matched by length of name.  So if you have two
    variables "a" and "ab", we'll test to see if the variable is
    "ab" before "a".  We also handle bracing of variable names
    like ${a} and ${ab} which will guarantee unambiguosity.

  * Variable expansion is always done on user input before 
    commands are evaluted.  This means variables can be used as 
    arguments to any commands, etc.  It also means that if you
    want the variable actually in the command, you have to prepend
    another $.

  * Variable expansion happens again when certain expressions are
    evaluated, such as action triggers, alias expansion, etc.
    Essentially, when any stored value gets expanded its variables
    will get expanded (typically whenever tintinmode would do its 
    sole expansion.) 

  * Placement vars in actions and aliases support an expanded 
    syntax based on python array slices.  %0:1 will evaluate to 
    the first and second arguments, while %0: will be all 
    arguments, and %:-1 will be all but the last argumen, etc.

Examples:

1. #action {$Myvar} {woo} 

Will trigger whenever the value Myvar holds when this is entered 
passes by (the original value of Myvar - it will not change if 
Myvar's value changes)

2. #action {$$Myvar} {woo}

Will trigger whenever the current value of Myvar passes by.

3. #action {$$$Myvar} {woo}

Will trigger whenever the literal string $Myvar is seen.  Place 
more $s in if you wish to trigger on more of them, the first 2 
will be stripped by the variable expansion processes. 

4. #alias {hello} {$moo}

Will bind hello to $moo's current value.

5. #alias {hello} {$$moo}

Will bind hello to always expand to moo's current value at the
time of expansion.

6. #alias {hello} {$$$moo}

Will bind hello to expand to the literal string "$moo"

7. #alias {hello} {$${moo}}

Will bind hello to expand to moo's current value at the time of
expansion.

category: readme
"""

general = """
Type "#help help" for help on how to use the in-game help system.

Read through the "#help readme" topics.  These will help as they 
will walk you through how Lyntin works, how to get additional 
help, where to go for answers, and what to do if you find a bug.  
These are also exported into the README file.

You should read through the topics in "#help commands" for all 
the currently registered Lyntin commands.

Each user interface has its own help topic--these will be on the 
top level of the help structure.

To start, the "#session" command will allow you to start a 
session.  When you're done, "#end" will close Lyntin.

All documentation that comes with Lyntin is also available on the
Lyntin web-site.  If you find problems with the documentation or
have fixes for it, let us know on the lyntin-devl mailing list.

category: readme
"""

gettingstarted = """
Lyntin incorporates the _look and feel_ of Tintin, so if you've used
Tintin or a variant it should be pretty easy to make the transition.  
It offers all the major features of Tintin, including multiple 
sessions, but leaves some things out.

There are *in-game* help files covering the commands and most 
other things we could think of which can be accessed with the "#help"
command.

All the documentation comes with Lyntin and is also available on
the Lyntin website in HTML form.  If you find problems with the 
documentation or have fixes for it, let us know.

You can use a config file to configure Lyntin at startup.  Alternatively
you can edit the startup script and embed the boot options which
get passed into the lyntin.engine.main function.

category: readme
"""

osnotes = """
Lyntin works in most environments we can think of, but it has some
caveats with the various operating systems due to differences 
between them.

WINDOWS
-------

Windows users should use either \\ or / to denote directory 
separaters.

ex::

   #write c:\\commandfile
   #write c:/commandfile


REDHAT LINUX
------------

Depending on which version of RedHat Linux you have, you will have 
to install the Python RPM as well as the Python Tkinter RPM.  If 
you don't install the Tkinter RPM, then you won't be able to use 
the Tk ui and it'll complain that it's missing a library when you 
try.


MAC OSX
-------

I have no experience with Mac OSX but after reading the various 
pages on Python and how it works on OSX, I'm hesitant to say 
Lyntin is fully supported.  However, I don't know of any reason 
it shouldn't be supported either except that one person on the 
mailing list has had problems with getting #write to work.


OTHER NOTES
-----------

If you encounter other operating system issues, let us know both 
the problem and the solution so we can add them here.

category: readme
"""

regexp = """
Lyntin allows the use of regular expressions in various arguments
for commands like #action, #highlight, and such.  It uses a 
specific format to trigger using raw regular expressions rather 
than having your argument get escaped so it can be compiled into 
a regular expression.  This allows you to write arguments using 
the simplest form that you can without having to adjust toggles 
and such.

For example:

  #highlight {red} {*says:}

is the same as:

  #highlight {red} {r[^.*?says:]}

The first one will undergo escaping and get transformed into 
"^.*?says\:" (without the quotes) before being compiled into a
regular expression.

The second one gets compiled without being escaped.

If you want to pass an "ignorecase" flag, do so after the end
]:

  #highlight {red} {r[krynor]i}

will highlight all instances of "krynor" (ignoring case) as red.

For regular expression documentation, refer to the Python 
documentation at:

  http://www.python.org/doc/current/lib/re-syntax.html

Note: It may have moved since this was written.

category: readme
"""

whylyntin = """
Lyntin is written entirely in Python--a nicely written and very 
portable programming language.  Thusly Lyntin is platform 
independent.  Lyntin is exposes the Python interpreter allow you 
more freedom than mere 'if you see this then send this' aliases 
and triggers.  They can be Python functions which do anything 
from setting a variable to forking a web spider.  In addition, 
your code can interface with Lyntin's code.  

Lyntin is great if:

1. you want a mud client that you can see the source code to
   and adjust it to suit your needs
2. you want a mud client that has a sophisticated API for 
   enhancing and building bots/agents/triggers more advanced than
   "if you see this then do this"
3. you want a mud client that works on all your machines,
   has a text ui, a tk gui, and can also work over ssh/telnet

Lyntin is not great if:

1. you prefer wizards and menus to the command line
2. you hate Python
3. you want fancy bells and whistles in the ui

category: readme
"""

def load():
  exported.add_help("bugs", bugs)
  exported.add_help("command", command)
  exported.add_help("contribute", contribute)
  exported.add_help("errata", errata)
  exported.add_help("evaluation", evaluation)
  exported.add_help("general", general)
  exported.add_help("gettingstarted", gettingstarted)
  exported.add_help("osnotes", osnotes)
  exported.add_help("regexp", regexp)
  exported.add_help("whylyntin", whylyntin)



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: advanced.py,v 1.9 2007/09/30 02:02:50 willhelm Exp $
#########################################################################
"""
This module holds the magical python_cmd code.  It takes in code,
and attempts to execute it in the lyntinuser.py module.  If no such
module exists, it executes it in this module.

It also holds load_cmd which does a lot of other magic stuff.
"""
import sys
import io
from code import compile_command
from lyntin import exported, config

usermodule = None
execdictglobals = None
execdictlocals = None

def _get_user_module():
  """
  Imports and returns the nicest user module it can find.  If we've
  already loaded a usermodule, then we use the cached one we
  loaded before so we're not doing this over and over again.

  @returns: the user module we just loaded or None
  @rtype: module
  """
  global usermodule
  if usermodule:
    return usermodule

  # this probably isn't exactly right since it'll look for the
  # first "lyntinuser" it finds and use that one.  i'm not sure how
  # we could implement a priority.
  for mem in sys.modules.keys():
    modname = "lyntinuser"
    if mem == modname or mem.endswith("." + modname):
      return sys.modules[mem]

  return None


def python_cmd(ses, words, input):
  """
  #@ allows you to execute arbitrary Python code inside of Lyntin.
  It will first look for a module named "lyntinuser" and execute
  the code inside that module's __dict__ environment.  If no
  such module exists, it will execute the code inside 
  modules.advanced .  At present it can only handle one-line
  Python statements.

  examples:
    #@ print "hello"
    #@ print "\\n".join(exported.get_commands())

  category: commands
  """
  global execdictglobals, execdictlocals

  # NOTE: if we ever get to handling multiple-lines, we'll need
  # to change this function completely.

  try:
    if execdictlocals == None:
      execdictlocals = {}
      
    execdictlocals["session"] = ses
    execdictlocals["exported"] = exported

    my_usermodule = _get_user_module() 
    if my_usermodule:
      dictglobals = my_usermodule.__dict__
    else:
      if execdictglobals == None:
        execdictglobals = {}
        exported.write_error("No lyntinuser module loaded--executing with no context.")
      dictglobals = execdictglobals

    source = input[1:].lstrip()
    compiled = compile_command(source)

    #
    # XXX for one-liners only:
    #
    if not compiled:
      compiled = compile_command(source+"\n")

    old_stdout = sys.stdout
    old_stderr = sys.stderr
    old_stdin = sys.stdin
    sys_stdout = io.StringIO()
    sys_stderr = io.StringIO()
    sys_stdin = io.StringIO()
    try:
      sys.stdout = sys_stdout
      sys.stderr = sys_stderr
      sys.stdin = sys_stdin
      
      #exec(compiled in dictglobals, execdictlocals)
      exec(compiled, dictglobals, execdictlocals)

    finally:
      sys.stdout = old_stdout
      sys.stderr = old_stderr
      sys.stdin = old_stdin
      
    error = sys_stderr.getvalue()
    if error:
      exported.write_error(error)

    text = sys_stdout.getvalue()
    if text.endswith("\n"):
      text = text[:-1]
    if text:  
      exported.write_message(text)  

  except (OverflowError, SyntaxError, ValueError, NameError):
    import traceback
    exported.write_error("".join(traceback.format_exception_only( *(sys.exc_info()[:2]) )))

  except:
    exported.write_traceback("@: error in raw python stuff.")
    exported.tally_error()


def load_cmd(ses, args, input):
  """
  Loads/reloads a module.

  When reloading, it looks for an "unload" function and executes 
  it prior to reloading the module.  After reloading/loading, it 
  looks for a "load" function and executes it.

  Lyntin modules located in the modules package are safe to reload 
  in-game.  Lyntin core modules (engine, helpmanager, event...) are
  NOT safe to import in-game.

  examples:
    #load modules.action
    #load exportuser

  #load will look for the module on the sys.path.  So if your module
  is not on the sys.path, you should first add the directory using #@:

    #@ import sys
    #@ sys.path.append("/directory/where/my/module/exists")

  Directories specified by the moduledir command-line argument are
  added to the sys.path upon Lyntin startup.

  category: commands
  """
  mod = args["modulename"]
  reload = args["reload"]

  # if this module has previously been loaded, we try to reload it.
  if mod in sys.modules:

    _module = sys.modules[mod]
    _oldmodule = _module

    try:
      if "lyntin_import" in _module.__dict__:
        # if we're told not to reload it, we toss up a message and then
        # do nothing
        if not reload:
          exported.write_message("load: module %s has already been loaded." % mod)
          return

        # if we loaded it via a lyntin_import mechanism and it has an
        # unload method, then we try calling that
        if "unload" in _module.__dict__:
          try:
            _module.unload()
          except Exception as e:
            exported.write_traceback(f"load: module {mod} didn't unload properly: {e}")
      del sys.modules[mod]
      exported.write_message("load: reloading %s." % mod)

    except:
      exported.write_traceback("load: had problems unloading %s." % mod)
      return
  else:
    _oldmodule = None


  # here's where we import the module
  try:
    _module = __import__( mod )
    _module = sys.modules[mod]

    if (_oldmodule and "reload" in _oldmodule.__dict__):
      try:
        _oldmodule.reload()
      except:
        exported.write_traceback("load: had problems calling reload on %s." % mod)
    
    if ("load" in _module.__dict__):
      _module.load()

    _module.__dict__["lyntin_import"] = 1
    exported.write_message("load successful.")
    if mod not in config.lyntinmodules:
      config.lyntinmodules.append(mod)

  except:
    exported.write_traceback("load: had problems with %s." % mod)

def unload_cmd(ses, args, input):
  """
  Unloads a module from Lyntin by calling the module's "unload" function
  and then removing references to it in the Python environment.

  examples:
    #unload wbgscheduler
    #unload modules.alias

  category: commands
  """
  mod = args["modulename"]

  if mod in sys.modules:
    _module = sys.modules[mod]

    if "lyntin_import" in _module.__dict__:
      if "unload" in _module.__dict__:
        try:
          _module.unload()
        except:
          exported.write_traceback("unload: module %s didn't unload properly." % mod)
      else:
        exported.write_error("unload: module %s doesn't have an unload function." % mod)

      del sys.modules[mod]
      exported.write_message("unload: module %s unloaded." % mod)
      if mod in config.lyntinmodules:
          config.lyntinmodules.remove(mod)
      return
    else:
      exported.write_error("unload: module %s cannot be unloaded." % mod)
      return

  exported.write_error("unload: module %s is not loaded." % mod)


def load():
  exported.add_command("@", python_cmd)
  exported.add_command("^load", load_cmd, "modulename reload:boolean=true")
  exported.add_command("^unload", unload_cmd, "modulename")

def unload():
  exported.remove_command("@")
  exported.remove_command("^load")
  exported.remove_command("^unload")

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
# 
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# 
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
# copyright (c) Free Software Foundation 2001-2007
# 
# $Id: ansi.py,v 1.2 2007/07/24 00:39:03 drexl Exp $
#########################################################################
"""
This holds a series of classes and functions for helping to manipulate
ANSI color codes.

In general, Lyntin keeps the data from the mud intact without doing any
transformations on it letting the ui do the transformations it needs to
do to display the mud data.  The exception to this is when the user has
shut off mudansi using the #config command.  Then we'll whack any incoming
ANSI color stuff before moving it around.
"""
import re

# for finding ANSI color sequences
ANSI_COLOR_REGEXP = re.compile(chr(27) + '\[[0-9;]*[m]')

STYLE_NORMAL = 0
STYLE_BOLD = 1
STYLE_UNDERLINE = 4
STYLE_BLINK = 5
STYLE_REVERSE = 7

# enums for placement in the color list
PLACE_BOLD = 0
PLACE_UNDERLINE = 1
PLACE_BLINK = 2
PLACE_REVERSE = 3
PLACE_FG = 4
PLACE_BG = 5

# the default color
DEFAULT_COLOR = [0, 0, 0, 0, -1, -1]

# used for converting text descriptions to ANSI color sequences
STYLEMAP = {
             "default": "0",
             "bold": "1",
             "underline": "4",
             "blink": "5",
             "reverse": "7",
             "black": "30",
             "red": "31",
             "green": "32",
             "yellow": "33",
             "blue": "34",
             "magenta": "35",
             "cyan": "36",
             "white": "37",
             "grey": "1;30",
             "light red": "1;31",
             "light green": "1;32",
             "light yellow": "1;33",
             "light blue": "1;34",
             "light magenta": "1;35",
             "light cyan": "1;36",
             "light white": "1;37",
             "b black": "40",
             "b red": "41", 
             "b green": "42",
             "b yellow": "43",
             "b blue": "44",
             "b magenta": "45",
             "b cyan": "46",
             "b white": "47"
           }


def filter_ansi(text):
  """
  Takes in text and filters out the ANSI color codes.

  @returns: text without ANSI color codes
  @rtype: string
  """
  return ANSI_COLOR_REGEXP.sub('', text)


def is_color_token(token):
  """
  Returns whether or not this is a color token.  It figures this out
  by checking to see if the token matches this regexp: 
  chr(27) + '\[[0-9;]*[m]'

  @param token: the token in question
  @type  token: string

  @return: 1 if it's color, 0 if not
  @rtype: boolean
  """
  if len(token) == 0:
    return 0

  return ANSI_COLOR_REGEXP.match(token)


def fix_color(color):
  """
  Helper function for debugging--it'll fix a color token so it's 
  readable in ascii.  It just replaces instances of chr(27) with 
  "ESC".

  @param color: the color token
  @type  color: string

  @return: the pretty string
  @rtype: string
  """
  return color.replace(chr(27), "ESC")


def split_ansi_from_text(text):
  """
  Takes in a string and separates it into a list of strings and ansi
  color strings.

  @param text: the full string to split up
  @type  text: string

  @return: list of text and ansi color tokens
  @rtype: list of strings
  """
  global ANSI_COLOR_REGEXP

  matchob = ANSI_COLOR_REGEXP.search(text)
  if matchob:
    textlist = []
    marker = 0
    while matchob:
      (b, e) = matchob.span()

      if marker == b:
        textlist.append(text[b:e])
      else:
        textlist.append(text[marker:b])
        textlist.append(text[b:e])

      marker = e
      matchob = ANSI_COLOR_REGEXP.search(text, marker)

    # we do this to handle ansi color sequences which are broken
    # between two network chunks
    if marker < len(text):
      escindex = text.rfind('\33', marker)
      if escindex != -1:
        for i in range(escindex+1, len(text)):
          c = text[i]

          if c.isdigit() or c == ";" or c == "[":
            continue
          else:
            escindex = -1
            break

      if escindex == -1:
        textlist.append(text[marker:])
      else:
        textlist.append(text[marker:escindex])
        textlist.append(text[escindex:])

    return textlist
  return [text]


def figure_color(textlist, currentcolor, leftover=""):
  """ 
  Takes a textlist of text and color tokens and figures out
  the latest current color.

  @param textlist: the list of string and ansi color code tokens
  @type  textlist: list of strings

  @param currentcolor: a tuple of (options, foreground, background) 
      that represent the current color
  @type  currentcolor: (int, int, int)

  @param leftover: if we encounter a half done color code, we throw
      it in the leftover string.  the leftover gets prepended
      to the textlist element on the next run of figure_color
  @type  leftover: string

  @return: the new currentcolor and leftover as a tuple
  @rtype: ((int, int, int), string)
  """
  global ANSI_COLOR_REGEXP, DEFAULT_COLOR

  if type(textlist) == type(''):
    textlist = split_ansi_from_text(textlist)

  if leftover:
    first = leftover + textlist[0]
    matchob = ANSI_COLOR_REGEXP.search(first)
    if matchob:
      (b, e) = matchob.span()
      textlist.insert(0, first[:e])
      textlist[1] = first[e:]
    leftover = ''

  for color in textlist:
    if is_color_token(color):
      color = color[2:-1]

      # handles the case where it's ESC[m which is short-hand for ESC[0m
      if color == "":
        currentcolor = list(DEFAULT_COLOR)

      # handles other cases!
      else:
        color = color.split(";")
        for i in color:
          if not i.isdigit():
            continue

          i = int(i)

          if i == 0:
            # 0 is a reset
            currentcolor = list(DEFAULT_COLOR)
      
          elif i == 1:
            currentcolor[PLACE_BOLD] = 1

          elif i == 4:
            currentcolor[PLACE_UNDERLINE] = 1

          elif i == 5:
            currentcolor[PLACE_BLINK] = 1

          elif i == 7:
            currentcolor[PLACE_REVERSE] = 1


          elif i == 22:
            currentcolor[PLACE_BOLD] = 0

          elif i == 24:
            currentcolor[PLACE_UNDERLINE] = 0

          elif i == 25:
            currentcolor[PLACE_BLINK] = 0

          elif i == 27:
            currentcolor[PLACE_REVERSE] = 0


          elif i == 39:
            # sets default foreground
            currentcolor[PLACE_FG] = -1

          elif 30 <= i and i <= 37:
            # these are foreground attributes
            currentcolor[PLACE_FG] = i

          elif i == 49:
            # sets default background
            currentcolor[PLACE_BG] = -1

          elif 40 <= i and i <= 47:
            # these are background attributes
            currentcolor[PLACE_BG] = i

  # we're looking for leftover pieces here
  if len(textlist) > 0:
    mem = textlist[-1]
    esc = mem.find('\33')
    if esc != -1:
      for i in range(esc, len(mem)):
        c = mem[i]

        if c.isdigit() or c == ";" or c == "[":
          continue
        else:
          esc = -1

      if esc != -1:
        leftover = mem
      
  return currentcolor, leftover


def get_color(style):
  """
  Looks at the style (which is a comma separated list of 
  styles) and figures out the markup string and returns it.

  @param style: the style to retrieve markup for
  @type  style: text

  @return: the ansi code markup for the given style
  @rtype: string
  """
  styles = style.split(",")
  markup = ""
  for mem in styles:
    mem = mem.strip()
    if mem in STYLEMAP:
      markup = markup + STYLEMAP[mem] + ";"
  return chr(27) + "[" + markup[:-1] + "m"


def convert_tuple_to_ansi(token):
  """
  Takes in a color tuple like what figure_color creates
  and converts it into an ANSI color sequence.

  @param token: the color tuple
  @type  token: tuple of ints

  @return: the ANSI color string
  @rtype: string
  """
  color = []

  if token[PLACE_BOLD] == 1:
    color.append("1")

  if token[PLACE_UNDERLINE] == 1:
    color.append("4")

  if token[PLACE_BLINK] == 1:
    color.append("5")

  if token[PLACE_REVERSE] == 1:
    color.append("7")

  if token[PLACE_FG] != -1:
    color.append(str(token[PLACE_FG]))

  if token[PLACE_BG] != -1:
    color.append(str(token[PLACE_BG]))

  if len(color) == 0:
    return chr(27) + "[0m"

  return chr(27) + "[" + ";".join(color) + "m"

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:
